#include "mainobserver.h"



typedef std::vector<double> Vec;
typedef std::vector<Vec> Mat;


Vec operator*(const Mat &a, const Vec &x) {
	int i, j;
	int m = a.size();
	int n = x.size();

	Vec prod(m);

	for (i = 0; i < m; i++) {
		prod[i] = 0.;
		for (j = 0; j < n; j++)
			prod[i] += a[i][j] * x[j];
	}
	return prod;
}


auto Flatland::Player::updatepastlightcone(Mainworldobserver* mainobserver) -> void
{
	for (auto x = 0; x < 80; ++x)
	{
		pastlightcone[x] = '-';
	}

	auto hmm = std::vector<std::tuple<Flatlandobject*, double>>(mainobserver->objectlist.begin() + 1, mainobserver->objectlist.end());

	for (auto object : hmm)
	{
		auto theobject = std::get<0>(object);

		for (auto it = theobject->objecthistory.begin(); it != theobject->objecthistory.end(); ++it)
		{
			auto objectstate = *it;

			auto mainclocktimeatstatechange = std::get<1>(objectstate);
			auto propertimeclockatstatechange = std::get<2>(objectstate);
			auto maincoordinatelocationatstatechange = std::get<3>(objectstate);
			auto mainrelativevelocity = std::get<4>(objectstate);
			auto properacceleration = std::get<5>(objectstate);

			if (std::get<0>(objectstate) == INERTIAL)
			{
				auto lightconeeq1 = std::vector<double>{ 1.0, (mainobserver->currenttime*LIGHTSPEED - std::get<1>(mainobserver->objectlist[0])) };
				auto lightconeeq2 = std::vector<double>{ -1.0, (mainobserver->currenttime*LIGHTSPEED + std::get<1>(mainobserver->objectlist[0])) };
				auto worldlineeq = std::vector<double>{ LIGHTSPEED/ mainrelativevelocity , (mainclocktimeatstatechange*LIGHTSPEED - maincoordinatelocationatstatechange* (LIGHTSPEED / mainrelativevelocity))};

				//get intersection points(always exists since v<c)
				auto intersection1 = std::vector<double>{
					(lightconeeq1[0] * worldlineeq[1] - worldlineeq[0] * lightconeeq1[1]) / (lightconeeq1[0] - worldlineeq[0]),
					(worldlineeq[1] - lightconeeq1[1]) / (lightconeeq1[0] - worldlineeq[0]),
				};
				auto intersection2 = std::vector<double>{
					(lightconeeq2[0] * worldlineeq[1] - worldlineeq[0] * lightconeeq2[1]) / (lightconeeq2[0] - worldlineeq[0]),
					(worldlineeq[1] - lightconeeq2[1]) / (lightconeeq2[0] - worldlineeq[0]),
				};

				//auto & nextobjectstate = *(std::next(it));

				auto lasttime = mainobserver->currenttime;

				if (std::next(it) != theobject->objecthistory.end())
				{
					auto & nextobjectstate = *(std::next(it));
					lasttime = std::get<1>(nextobjectstate);
				}

				if (mainclocktimeatstatechange <= intersection1[0] / LIGHTSPEED && intersection1[0] / LIGHTSPEED < lasttime)
				{
					intersection1 = std::vector<double>{ intersection1[0] - LIGHTSPEED * mainobserver->currenttime, intersection1[1] - std::get<1>(mainobserver->objectlist[0]) };
					intersection1 = LORENTZ_BOOST(-currentvelocity)*intersection1;

					intersection1[0] = intersection1[0] / LIGHTSPEED;


					if (-40.0 < intersection1[1] && intersection1[1] < 40.0)
					{
						pastlightcone[(int)round(intersection1[1])] = theobject->name;
					}
				}
				if (mainclocktimeatstatechange <= intersection2[0] / LIGHTSPEED && intersection2[0] / LIGHTSPEED < lasttime)
				{
					intersection2 = std::vector<double>{ intersection2[0] - LIGHTSPEED * mainobserver->currenttime, intersection2[1] - std::get<1>(mainobserver->objectlist[0]) };
					intersection2 = LORENTZ_BOOST(-currentvelocity)*intersection2;

					intersection2[0] = intersection2[0] / LIGHTSPEED;


					if (-40.0 < intersection2[1] && intersection2[1] < 40.0)
					{
						pastlightcone[(int)round(intersection2[1])+40] = theobject->name;
					}
				}
			}
			else
			{
				auto lightconeeq1 = std::vector<double>{ 1.0, (mainobserver->currenttime*LIGHTSPEED - std::get<1>(mainobserver->objectlist[0])) };
				auto lightconeeq2 = std::vector<double>{ -1.0, (mainobserver->currenttime*LIGHTSPEED + std::get<1>(mainobserver->objectlist[0])) };

				auto intersection1 = std::vector<double>{ 0,0 };
				auto intersection2 = std::vector<double>{ 0,0 };

				// point at which mainrelativevelocity is tangent to the acceleration hyperbola (t', x') with (0,0) being v=0
				// so (t', x') -> (mainclocktimeatstatechange, maincoordinatelocationatstatechange)
				auto accelerationeqoffset = std::vector<double>{
					(LIGHTSPEED*GAMMA(mainrelativevelocity)*mainrelativevelocity / properacceleration) - LIGHTSPEED * mainclocktimeatstatechange,
					((LIGHTSPEED*LIGHTSPEED / properacceleration)*(sqrt((GAMMA(mainrelativevelocity)*mainrelativevelocity / LIGHTSPEED)*(GAMMA(mainrelativevelocity)*mainrelativevelocity / LIGHTSPEED) + 1.0) - 1.0))  - maincoordinatelocationatstatechange };

				//change lightcones eq. coord. so that they share (0,0) with acceleration hyperbola with (0,0) being v=0
				lightconeeq1[1] += accelerationeqoffset[0] - accelerationeqoffset[1];
				lightconeeq2[1] += accelerationeqoffset[0] + accelerationeqoffset[1];
					
				//change lightcones eq. coord. such that they now share (0,0) with shifted acceleration hyperbola with (0,c^2/a) being v=0
				lightconeeq1[1] += (LIGHTSPEED * LIGHTSPEED / properacceleration);
				lightconeeq2[1] += (-LIGHTSPEED * LIGHTSPEED / properacceleration);

				auto lasttime = mainobserver->currenttime;

				if (std::next(it) != theobject->objecthistory.end())
				{
					auto & nextobjectstate = *(std::next(it));
					lasttime = std::get<1>(nextobjectstate);
				}

				if (lightconeeq1[1]*lightconeeq1[1] > 0.0) // determinant for intersection
				{

					intersection1[1] = ((LIGHTSPEED*LIGHTSPEED*LIGHTSPEED*LIGHTSPEED) / (properacceleration*properacceleration) + (lightconeeq1[1] * lightconeeq1[1])) / (2 * lightconeeq1[1]);
					intersection1[0] = intersection1[1] + lightconeeq1[1];

					//shift intersection point x back to main observer coordinates
					intersection1[1] -= (accelerationeqoffset[1]) + ((LIGHTSPEED * LIGHTSPEED) / properacceleration);
					//shift intersection point y back to main observer coordinates
					intersection1[0] -= (accelerationeqoffset[0]);
					
					
					if (mainclocktimeatstatechange <= intersection1[0] / LIGHTSPEED && intersection1[0] / LIGHTSPEED < lasttime)
					{
						intersection1 = std::vector<double>{ intersection1[0] - LIGHTSPEED * mainobserver->currenttime, intersection1[1] - std::get<1>(mainobserver->objectlist[0]) };
						intersection1 = LORENTZ_BOOST(-currentvelocity)*intersection1;

						intersection1[0] = intersection1[0] / LIGHTSPEED;


						if (-40.0 < intersection1[1] && intersection1[1] < 40.0)
						{
							pastlightcone[(int)round(intersection1[1])] = theobject->name;
						}
					}

				}
				else if (lightconeeq2[1] * lightconeeq2[1] > 0.0)
				{
					intersection2[1] = ((LIGHTSPEED*LIGHTSPEED*LIGHTSPEED*LIGHTSPEED) / (properacceleration*properacceleration) + (lightconeeq2[1] * lightconeeq2[1])) / (2.0 * lightconeeq2[1]);
					intersection2[0] = intersection2[1] + lightconeeq2[1];
						
					//shift intersection point x back to main observer coordinates
					intersection2[1] -= (accelerationeqoffset[1]) - ((LIGHTSPEED * LIGHTSPEED) / properacceleration);
					//shift intersection point y back to main observer coordinates
					intersection2[0] -= (accelerationeqoffset[0]);

					if (mainclocktimeatstatechange <= intersection2[0] / LIGHTSPEED && intersection2[0] / LIGHTSPEED < lasttime)
					{
						intersection2 = std::vector<double>{ intersection2[0] - LIGHTSPEED * mainobserver->currenttime, intersection2[1] - std::get<1>(mainobserver->objectlist[0]) };
						intersection2 = LORENTZ_BOOST(-currentvelocity)*intersection2;

						intersection2[0] = intersection2[0] / LIGHTSPEED;


						if (-40.0 < intersection2[1] && intersection2[1] < 40.0)
						{
							pastlightcone[(int)round(intersection2[1])] = theobject->name;
						}
					}
				}
			}
		}

		//for (auto objectstate : theobject.objecthistory)
		//{
		//	
		//}
	}

	return;
}


auto Flatland::Player::changestate(double maintime) -> void
{
	return;
}




auto Flatland::Object1::changestate(double maintime) -> void
{
	if (maintime / 5.0 >= cnt)
	{
		for (auto i = cnt; i <= (maintime / 5.0); ++i)
		{
			auto v = 2.5 / sqrt(1.0+(2.5/LIGHTSPEED)*(2.5/LIGHTSPEED));
			Flatlandobject::objecthistory.push_back(std::make_tuple(ACCELERATING, cnt*5, GAMMA(5.0) * 5.0 * i, 5.0, (toggle)? -v:v, (toggle) ? 1 : -1));
			toggle = !toggle;
			cnt++;
		}
	}
	return;
}


auto Flatland::Mainworldobserver::updatebytick(double tick) -> void
{
	currenttime += tick;
	auto cnt = 0;
	for (std::tuple<Flatlandobject*, double> a : objectlist)
	{
		auto theobject = std::get<0>(a);
		
		
		auto objectstate = theobject->objecthistory.back();
		auto mainclocktimeatstatechange = std::get<1>(objectstate);
		auto propertimeclockatstatechange = std::get<2>(objectstate);
		auto maincoordinatelocationatstatechange = std::get<3>(objectstate);
		auto mainrelativevelocity = std::get<4>(objectstate);
		auto properacceleration = std::get<5>(objectstate);
		double newx;
		/*if (std::get<0>(objectstate) == INERTIAL)
		{
			newx = maincoordinatelocationatstatechange + mainrelativevelocity * (this->currenttime - mainclocktimeatstatechange);
			std::tuple<Flatlandobject, double> newobjectinfo = std::make_tuple(theobject, newx);
			a.swap(newobjectinfo);
			free(&newobjectinfo);
			auto a = std::vector<double>{this->currenttime - mainclocktimeatstatechange, newx- maincoordinatelocationatstatechange};
			a = LORENTZ_BOOST(theobject.currentvelocity) * a;
			theobject.propertime = propertimeclockatstatechange + a[0];
			theobject.currentproperacceleration = 0;
		}
		else
		{

		}
		/*for (std::tuple<Flatlandobject, double> apart : (theobject.parts))
		{
			auto thepart = std::get<0>(apart);
			auto relativelocationtoCoM = std::get<1>(apart);
			auto partstate = thepart.objecthistory.back();
			if (std::get<0>(partstate) == INERTIAL)
			{
				auto a = std::vector<double>{ theobject.propertime, (double)relativelocationtoCoM };
				a = LORENTZ_BOOST(-theobject.currentvelocity) * a;
				a = std::vector<double>{ a[0] + this->currenttime , a[1] + newx };
				thepart.currentvelocity = theobject.currentvelocity;
			}
		}*/
		theobject->changestate(currenttime);
	}
}

